!error
! Module: esrcmerge_mod
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Merges two source lists (of either eboxdetect or emldetect output style) into one.
! 
! TODO:
!
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module esrcmerge_mod

  use dal
  use param  
  use errorHandling

  use array_utils	! for smallestTrueIndex
  use dal_aux		! for splitSetTabName

  implicit none 
  
  private 

  ! Command-line parameters:
  character(256), allocatable ::&
    inSrcListSetTabNames(:)
  character(256) ::&
    tempSetName = '',&
    outSrcListSetName = ''
  character(20) ::&
    idColName = ''
  integer(int8) ::&
    userIntNull = 0

!  type :: localColType
!    logical(bool),  pointer :: boolPtr(:)
!    integer(int8),  pointer :: int8Ptr(:)
!    integer(int16), pointer :: int16Ptr(:)
!    integer(int32), pointer :: int32Ptr(:)
!    real(single),   pointer :: singlePtr(:)
!    real(double),   pointer :: doublePtr(:)
!    type(ColumnT)           :: col
!    logical                 :: isNullable
!    integer                 :: stringLength
!  end type localColType

  type :: colDataType
    character(256) :: name, units, comment
    integer(int32) :: dataType
    integer :: dims(0:0)
  end type colDataType


  public esrcmerge

contains 
  
  !---------------------------------------------------------------------
  subroutine esrcmerge
  !---------------------------------------------------------------------

    implicit none

    integer :: numInSets, setNum, numInCols, rowStart, colNum, totalNumCols, i
    integer, allocatable :: numRows(:), numCols(:)
    character(256), allocatable :: setNames(:)
    character(256) :: colName
    character(8), allocatable :: tabNames(:)
    type(DataSetT) :: inSet, outSet
    type(TableT) :: inTab, outTab
    type(ColumnT) :: inCol, outCol
    integer, pointer :: inDims(:)
    type(colDataType), allocatable :: colData(:)

    !-------------------------------------------------------------------
    ! Start


    call readParameters

    numInSets = size(inSrcListSetTabNames)
    if (numInSets <= 0) then
      call warning('',"")
  return
    else if (numInSets == 1) then
      call warning('',"")
!*** copy it to outSrcListSetName
  return
    end if

    allocate(numRows(numInSets),  numCols(numInSets))
    allocate(setNames(numInSets), tabNames(numInSets))

    ! Get the set and tab names of the input files and their numbers of rows:

    do setNum = 1, numInSets
      call splitSetTabName(inSrcListSetTabNames(setNum), setNames(setNum)&
        , tabNames(setNum))
      inSet = dataset(setNames(setNum), READ)
      inTab = table(inSet, tabNames(setNum))

      if (setNum == 1 .and. .not.hasColumn(inTab, idColName))&
call error('noIdColInFirstTable',"")

      inCol = column(inTab, idColName)
      inDims => dimensions(inCol)
if (size(inDims) .ne. 1) call error('badColumnDimension',"")
if (inDims(0) .ne. 1) call error('vectorIdColumn',"")
      call release(inCol)

      numRows(setNum) = numberOfRows(inTab)
      numCols(setNum) = numberOfColumns(inTab)
      call release(inTab)
      call release(inSet)
    end do

    ! Now set up the column names, data types etc of the complete list
    ! of columns:

    allocate(colData(sum(numCols))) ! stupid, but since an attempt to
    ! do it nicely by use of pointers makes the program seg fault....

    call initializeColData(colData)

    totalNumCols = 0
    do setNum = 1, numInSets
      call splitSetTabName(inSrcListSetTabNames(setNum), setNames(setNum)&
        , tabNames(setNum))
      inSet = dataset(setNames(setNum), READ)
      inTab = table(inSet, tabNames(setNum))

      call addColData(inTab, colData, totalNumCols)

      call release(inTab)
      call release(inSet)
    end do

    outSet = dataSet(outSrcListSetName, CREATE)
    outTab = addTable(outSet, tabNames(1), sum(numRows))

    inSet = dataset(setNames(1), READ)
    call setAttributes(attributable(outSet), attributable(inSet))
    inTab = table(inSet, tabNames(1))
    call setAttributes(attributable(outTab), attributable(inTab))
    call release(inTab)
    call release(inSet)

    do i = 1, totalNumCols
      outCol = addColumn(outTab, trim(colData(i)%name), colData(i)%dataType&
        , colData(i)%units, (/colData(i)%dims(0)/))
      call release(outCol)
    end do

    call nullifyTable(outTab)

    rowStart = 0
    do setNum = 1, numInSets
      inSet = dataset(setNames(setNum), READ)
      inTab = table(inSet, tabNames(setNum))
      numInCols = numberOfColumns(inTab)

      do colNum = 0, numInCols-1
        inCol = column(inTab, colNum)
        colName = name(inCol)
        outCol = column(outTab, colName)
        if (colName .eq. idColName) then
          call transferIndexColData(inCol, rowStart, outCol)
        else
          call transferColData(inCol, rowStart, outCol)
        end if
        call release(outCol)
        call release(inCol)
      end do
      call release(inTab)
      call release(inSet)

      rowStart = rowStart + numRows(setNum)
    end do
    call release(outTab)
    call release(outSet)

  end subroutine esrcmerge
  
  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    integer :: numInSets, setNum

    !-------------------------------------------------------------------
    ! Start

    numInSets = parameterCount('insrclisttabs')
  if (numInSets <= 0) return
    allocate(inSrcListSetTabNames(numInSets))
    do setNum = 1, numInSets
      inSrcListSetTabNames(setNum) = stringParameter('insrclisttabs', setNum-1)
    end do

    tempSetName = stringParameter('tempset')
    outSrcListSetName = stringParameter('outsrclistset')
    idColName = stringParameter('idcol')
    userIntNull = intParameter('intnull')

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine initializeColData(colData)
  !---------------------------------------------------------------------

    implicit none

    type(colDataType), intent(out) :: colData(:)

    !-------------------------------------------------------------------
    ! Start

    colData%name     = ''
    colData%units    = ''
    colData%comment  = ''
    colData%dataType = 0
    colData%dims(0)  = 0

  end subroutine initializeColData

  !---------------------------------------------------------------------
  subroutine addColData(tab, colData, totalNumCols)
  !---------------------------------------------------------------------

    implicit none

    type(TableT),      intent(in)    :: tab
    type(colDataType), intent(inout) :: colData(:)
    integer,           intent(inout) :: totalNumCols

    integer :: numCols, i, oldColNum
    type(ColumnT) :: col
    character(256) :: colName
    integer(int32) :: datType
    integer, pointer :: dims(:)

    !-------------------------------------------------------------------
    ! Start

    numCols = numberOfColumns(tab)
    do i = 0, numCols-1
      col = column(tab, i)
      colName = name(col)
print *, 'colName=', trim(colName)
      datType = dataType(col)
print *, 'datType=', datType
      dims => dimensions(col)

      if (elementIsInList(colName, colData(:totalNumCols)%name)) then
      ! check that all else is the same:
        oldColNum&
          = smallestTrueIndex((colName .eq. colData(:totalNumCols)%name))
        if (colName .ne. colData(oldColNum)%name)&
call error('bug0',"Please contact the code developer.")

if (datType .ne. colData(oldColNum)%dataType) call error('wrongDataType',"")
if (size(dims) .ne. 1) call error('badColumnDimension',"")
print *, dims(0), colData(oldColNum)%dims(0)
if (dims(0) .ne. colData(oldColNum)%dims(0)) call error('wrongColumnVectorSize',"")
      else
        totalNumCols = totalNumCols + 1
        colData(totalNumCols)%name     = colName
        colData(totalNumCols)%dataType = datType
        colData(totalNumCols)%dims(0)  = dims(0)
        colData(totalNumCols)%units    = units(col)
!        colData(totalNumCols)%comment  = comment(col)
      end if

      call release(col)
    end do

  end subroutine addColData

  !---------------------------------------------------------------------
  logical function elementIsInList(str, strList)
  !---------------------------------------------------------------------
!*** put this in ssc_misc??

    implicit none

    character(*), intent(in) :: str, strList(:)

    integer :: numElements, i

    !-------------------------------------------------------------------
    ! Start

    elementIsInList = .false. ! default
    numElements = size(strList)
    do i = 1, numElements
      if (str .eq. strList(i)) then
        elementIsInList = .true.
  return
      end if
    end do

  end function elementIsInList

  !---------------------------------------------------------------------
  subroutine nullifyTable(tab)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(inout) :: tab

    integer :: numCols, numRows, i, row
    type(ColumnT) :: col
    integer, pointer :: dims(:)
    logical(bool), pointer :: boolPtr(:), boolVecPtr(:,:)
    logical :: isNullable
    character(256) :: dummyStr
!logical :: testFlag = .true.

    !-------------------------------------------------------------------
    ! Start

    numCols = numberOfColumns(tab)
    numRows = numberOfRows(tab)

    do i = 0, numCols-1
      col = column(tab, i)
      dims => dimensions(col)

      isNullable = .false. ! default

      select case(size(dims))
      case(1)
        if (dims(0) == 1 .or. dataType(col) .eq. STRING) then
          select case(dataType(col))
          case(BOOLEAN)
            boolPtr => boolData(col)
            boolPtr = .false.

          case(STRING)
            dummyStr = ''
            do row = 0, numRows-1
              call setStringCell(col, row, dummyStr)
            end do

          case(INTEGER8, INTEGER16, INTEGER32)
            call setNullValue(col, int(userIntNull))
            isNullable = .true.

          case(REAL32, REAL64)
            isNullable = .true.

          case default
call error('badDataType',"")
          end select

        else ! dims(0) > 1 .and. dataType(col) .neq. STRING
          select case(dataType(col))
          case(BOOLEAN)
            boolVecPtr => boolVectorData(col)
            boolVecPtr = .false.

          case(STRING)
            dummyStr = ''
            do row = 0, numRows-1
              call setStringCell(col, row, dummyStr)
            end do

          case(INTEGER8, INTEGER16, INTEGER32)
            call setNullValue(col, int(userIntNull))
            isNullable = .true.

          case(REAL32, REAL64)
            isNullable = .true.

          case default
call error('badDataType',"")
          end select
        end if

      case default
call error('badDataType',"")
      end select

      if (isNullable) then
        do row = 0, numRows-1
          call setToNull(col, row)
        end do
      end if

      call release(col)
    end do

  end subroutine nullifyTable

  !---------------------------------------------------------------------
  subroutine transferIndexColData(innCol, rowStart, outCol)
  !---------------------------------------------------------------------

    implicit none

    type(ColumnT), intent(in)    :: innCol
    integer,       intent(in)    :: rowStart
    type(ColumnT), intent(inout) :: outCol

    integer :: numNewRows
    integer(int32) :: outDataType, idOffset
    integer(int8),  pointer :: innInt8Ptr(:),  outInt8Ptr(:)
    integer(int16), pointer :: innInt16Ptr(:), outInt16Ptr(:)
    integer(int32), pointer :: innInt32Ptr(:), outInt32Ptr(:)

    !-------------------------------------------------------------------
    ! Start

    numNewRows = numberOfRows(innCol)

    if (nullDefined(innCol)) then
      if (hasNulls(innCol))&
call error('idColHasNulls',"")
      ! id col is not allowed to contain nulls.
    end if

    outDataType = dataType(outCol)

    idOffset = 0 ! default

    ! The index column must be integer (otherwise it doesn't make sense):
    select case(outDataType)
    case(INTEGER8)
      innInt8Ptr => int8Data(innCol)
      outInt8Ptr => int8Data(outCol)
      if (rowStart > 0) idOffset = 1 + maxval(outInt8Ptr(0:rowStart-1))&
        - minval(innInt8Ptr)

      outInt8Ptr( rowStart:rowStart+numNewRows-1) = innInt8Ptr + idOffset

    case(INTEGER16)
      innInt16Ptr => int16Data(innCol)
      outInt16Ptr => int16Data(outCol)
      if (rowStart > 0) idOffset = 1 + maxval(outInt16Ptr(0:rowStart-1))&
        - minval(innInt16Ptr)

      outInt16Ptr( rowStart:rowStart+numNewRows-1) = innInt16Ptr + idOffset

    case(INTEGER32)
      innInt32Ptr => int32Data(innCol)
      outInt32Ptr => int32Data(outCol)
      if (rowStart > 0) idOffset = 1 + maxval(outInt32Ptr(0:rowStart-1))&
        - minval(innInt32Ptr)

      outInt32Ptr( rowStart:rowStart+numNewRows-1) = innInt32Ptr + idOffset

    case(BOOLEAN, STRING, REAL32, REAL64)
call error('forbiddenIdColumnDataType',"")
    case default
call error('badDataType',"")
    end select

  end subroutine transferIndexColData

  !---------------------------------------------------------------------
  subroutine transferColData(innCol, rowStart, outCol)
  !---------------------------------------------------------------------

    implicit none

    type(ColumnT), intent(in)    :: innCol
    integer,       intent(in)    :: rowStart
    type(ColumnT), intent(inout) :: outCol

    integer :: numNewRows, row
!    type(ColumnT) :: outCol
    integer(int32) :: outDataType
    integer, pointer :: innDims(:), outDims(:)
!    logical, allocatable :: inColNulls(:), outColNulls(:)
    logical(bool), pointer ::&
      innBoolPtr(:), outBoolPtr(:), innBoolVecPtr(:,:), outBoolVecPtr(:,:)
    integer(int8),  pointer ::&
      innInt8Ptr(:),  outInt8Ptr(:), innInt8VecPtr(:,:),  outInt8VecPtr(:,:)
    integer(int16), pointer ::&
      innInt16Ptr(:), outInt16Ptr(:), innInt16VecPtr(:,:), outInt16VecPtr(:,:)
    integer(int32), pointer ::&
      innInt32Ptr(:), outInt32Ptr(:), innInt32VecPtr(:,:), outInt32VecPtr(:,:)
    real(single), pointer ::&
      innReal32Ptr(:), outReal32Ptr(:), innReal32VecPtr(:,:),&
      outReal32VecPtr(:,:)
    real(double), pointer ::&
      innReal64Ptr(:), outReal64Ptr(:), innReal64VecPtr(:,:),&
      outReal64VecPtr(:,:)

    !-------------------------------------------------------------------
    ! Start

    numNewRows = numberOfRows(innCol)

    outDataType = dataType(outCol)
    if (dataType(innCol) .ne. outDataType)&
call error('bug1',"Please contact the code developer.")

    innDims => dimensions(innCol)
    if (size(innDims) .ne. 1)&
 call error('bug2',"Please contact the code developer.")

    outDims => dimensions(outCol)
    if (size(outDims) .ne. 1)&
 call error('bug3',"Please contact the code developer.")

    if (outDims(0) == 1) then
      select case(outDataType)
      case(BOOLEAN)
        innBoolPtr => boolData(innCol)
        outBoolPtr => boolData(outCol)
        outBoolPtr(rowStart:rowStart+numNewRows-1) = innBoolPtr
      case(STRING)
        do row = 0, numNewRows-1
          call setStringCell(outCol, rowStart + row, stringCell(innCol, row))
        end do
      case(INTEGER8)
        innInt8Ptr => int8Data(innCol)
        outInt8Ptr => int8Data(outCol)
        outInt8Ptr(rowStart:rowStart+numNewRows-1) = innInt8Ptr
      case(INTEGER16)
        innInt16Ptr => int16Data(innCol)
        outInt16Ptr => int16Data(outCol)
        outInt16Ptr(rowStart:rowStart+numNewRows-1) = innInt16Ptr
      case(INTEGER32)
        innInt32Ptr => int32Data(innCol)
        outInt32Ptr => int32Data(outCol)
        outInt32Ptr(rowStart:rowStart+numNewRows-1) = innInt32Ptr
      case(REAL32)
        innReal32Ptr => real32Data(innCol)
        outReal32Ptr => real32Data(outCol)
        outReal32Ptr(rowStart:rowStart+numNewRows-1) = innReal32Ptr
      case(REAL64)
        innReal64Ptr => real64Data(innCol)
        outReal64Ptr => real64Data(outCol)
        outReal64Ptr(rowStart:rowStart+numNewRows-1) = innReal64Ptr
      case default
call error('badDataType',"")
      end select

    else
      select case(outDataType)
      case(BOOLEAN)
        innBoolVecPtr => boolVectorData(innCol)
        outBoolVecPtr => boolVectorData(outCol)
        outBoolVecPtr(:,rowStart:rowStart+numNewRows-1) = innBoolVecPtr
      case(STRING)
        do row = 0, numNewRows-1
          call setStringCell(outCol, rowStart + row, stringCell(innCol, row))
        end do
      case(INTEGER8)
        innInt8VecPtr => int8VectorData(innCol)
        outInt8VecPtr => int8VectorData(outCol)
        outInt8VecPtr(:,rowStart:rowStart+numNewRows-1) = innInt8VecPtr
      case(INTEGER16)
        innInt16VecPtr => int16VectorData(innCol)
        outInt16VecPtr => int16VectorData(outCol)
        outInt16VecPtr(:,rowStart:rowStart+numNewRows-1) = innInt16VecPtr
      case(INTEGER32)
        innInt32VecPtr => int32VectorData(innCol)
        outInt32VecPtr => int32VectorData(outCol)
        outInt32VecPtr(:,rowStart:rowStart+numNewRows-1) = innInt32VecPtr
      case(REAL32)
        innReal32VecPtr => real32VectorData(innCol)
        outReal32VecPtr => real32VectorData(outCol)
        outReal32VecPtr(:,rowStart:rowStart+numNewRows-1) = innReal32VecPtr
      case(REAL64)
        innReal64VecPtr => real64VectorData(innCol)
        outReal64VecPtr => real64VectorData(outCol)
        outReal64VecPtr(:,rowStart:rowStart+numNewRows-1) = innReal64VecPtr
      case default
call error('badDataType',"")
      end select
    end if

    if ( outDataType .eq. INTEGER8&
    .or. outDataType .eq. INTEGER16&
    .or. outDataType .eq. INTEGER32) then
      if (.not.nullDefined(outCol))&
call error('bug4', "Please contact the code developer.")

      if (nullDefined(innCol)) then
        if (hasNulls(innCol)) then
          do row = 0, numNewRows-1
            if (isNull(innCol, row)) call setToNull(outCol, rowStart+row)
          end do
        end if
      end if
    end if

  end subroutine transferColData

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module esrcmerge_mod
